<?php
/* Logaholic Web Analytics software             Copyright(c) 2005-2015 Logaholic B.V.
 *                                                               All rights Reserved.
 * This code is subject to the Logaholic license. Unauthorized copying is prohibited.
 * support@logaholic.com                         http://www.logaholic.com/License.txt
*/

# Set Labels for the TransAlarm Mysql Tables
define("TBL_TRANSALARM_DATA" , $mysqlprefix . "_logaholic_transalarm_data" );
define("TBL_TRANSALARM_LOG" , $mysqlprefix . "_logaholic_transalarm_log" );

define("_TRANSALARM", "Transalarm");

define("_TRANSALARM_SANDBOX", "Transalarm Sandbox");
define("_TRANSALARM_DASHBOARD", "Transalarm Dashboard");
define("_TRANSALARM_LOG_DETAIL", "Transalarm Log Detail");

@define("_TIMESTAMP", "Timestamp");
@define("_METRIC", "Metric");
@define("_EVENT", "Event");
@define("_TEST_START", "Test Start");
@define("_TEST_END", "Test End");
@define("_BASELINE", "Test End");
@define("_BASELINE", "Test End");
@define("_DENOMINATOR", "Denominator");
@define("_OBSERVED", "Observed");

include_once(logaholic_dir() . "includes/math.php");

Class TransAlarm{
	var $from = 0;								# defines the daterange from which timestamp to start
	var $to = 0;								# defines the daterange to which timestamp we end the search
	var $config = array("ResetDays" => 28);		# Config array.. Data from TransAlarmConfig.json are loaded in here
	var $alarmProfiles = array();				# Array with profile names that need an alarm to be send to

	var $events = array(
		1 => '1: Traffic Drop Alarm',			# Alarm Fired
		2 => '2: Ongoing Drop',					# Ongoing Alarm
		3 => '3: Traffic Spike Alarm',			# Alarm Active
		4 => '4: Ongoing Spike',				# All Ok
 		5 => '5: Create',						# Create
		6 => '6: Reset Baseline',				# Reset Baseline
		7 => '7: Inconclusive'					# Inconclusive
	);

	var $transactions = true;

	var $sandbox = false;
	var $debug = false;

	var $kpis = array();						# List of KPI's with id and url
	var $GammaAlphaConstant = 1.314250010345;	# This is used for calculation in the TestModule

	var $processData = array();					# The data that needs to be proccessed.
	var $metrics = array(); 					# Array with the metrics (visits,Transaction,download.zip)	
	var $sumOfMetrics = array(); 				# Array with the total values of each metric	

	var $totalHours = 0;						# Count of the total amount of hours in the processData

	var $lastDataDateHour = 0;					# Actual last timestamp in the processData
	var $firstDataDateHour = 0;					# Actual first timestamp in the processData
 
	var $transAlarmRecipientsFile = "";			# This file contains data over the profile users's email and special notes

	var $persistantDataTable = "";
	var $LogTable = "";

	function __CONSTRUCT(){
		global $profile, $logaUtm;

		if(empty($profile)){
			return;
		}

		$this->transactions = (empty($logaUtm)) ? false : true;

		# Check if the config file exists, if so load it in..
		if(file_exists("files/TransAlarmConfig.json")){
			$this->config = json_decode( file_get_contents("files/TransAlarmConfig.json") , true );
		}

		# Set the daterange from ($this->config['ResetDays']) amount of days ago to yesterday midnight
		$max = $profile->max_db_timestamp;
		$this->from = mktime(0,0,0, date("m", $max),date("d", $max) - $this->config['ResetDays'] ,date("Y", $max) );
		$this->to = mktime(23,59,59, date("m", $max),date("d", $max) - 1 ,date("Y", $max) );
		
		
		# Set the Recipients File location
		$this->transAlarmRecipientsFile = "files/ProfileData - TM2 Data.csv";

		# set default values for transactions and visits
		$this->sumOfMetrics["Visits"] = 0;
		if($this->transactions){
			$this->sumOfMetrics["Transactions"] = 0;
		}

		$this->persistantDataTable = TBL_TRANSALARM_DATA;
		$this->LogTable = TBL_TRANSALARM_LOG;

	}

	function SandBoxMode($mode){
		global $db;
		if($mode){
			$this->sandbox = true;
			$this->persistantDataTable = TBL_TRANSALARM_DATA . "_sandbox";
			$this->LogTable = TBL_TRANSALARM_LOG . "_sandbox";

			$db->Execute("TRUNCATE TABLE ". $this->persistantDataTable);
			$db->Execute("TRUNCATE TABLE ". $this->LogTable);
		} else {
			$this->sandbox = false;
			$this->persistantDataTable = TBL_TRANSALARM_DATA;
			$this->LogTable = TBL_TRANSALARM_LOG;
		}
	}

	function CreateTaDataTable(){
		global $db;
		$query = "CREATE TABLE ". $this->persistantDataTable ." ( ".
		  "`profile` varchar(255) NOT NULL, ".
		  "`metric` varchar(255) NOT NULL, ".
		  "`baseline` float DEFAULT NULL, ".
		  "`last0+` int(11) DEFAULT NULL, ".
		  "`last-` int(11) DEFAULT NULL, ".
		  "`regressors` text, ".
		  "`hash` char(32) NOT NULL, ".
		  "UNIQUE KEY `hash` (`hash`) ".
		") ENGINE=MyISAM DEFAULT CHARSET=utf8";

		$db->Execute($query);
	}

	function CreateTaEventLogTable(){
		global $db;
		$query = "CREATE TABLE ". $this->LogTable ." (".
			"`tiemstamp` INT(11) NOT NULL, ".
			"`event_type` VARCHAR(55) NOT NULL, ".
			"`profile` VARCHAR(255) NOT NULL, ".
			"`metric` VARCHAR(255) NOT NULL, ".
			"`start_date` int(11), ".
			"`end_date` int(11), ".
			"`baseline` FLOAT, ".
			"`ROPEboundary` FLOAT, ".
			"`ObserverdDenominator` int(255), ".
			"`ExpectedMetric` int(255), ".
			"`ObserverdMetric` int(255), ".
			"`ObserverdVisits` int(11), ".
			"`ErrorCertainty` FLOAT, ".
			"`posteriorHigh` FLOAT, ".
			"`ResetCertainty` FLOAT, ".
			"`posteriorLow` FLOAT, ".
            "key profile (profile), ".
            "key metric (metric) ".
			") ENGINE=MyISAM CHARSET=utf8";

		$db->Execute($query);
	}

	/*
	** Log an event to the $this->LogTable table
	** Given array must have the correct column name as key and value as value to set
	** ['profile' => test]
	*/
	function LogEvent($dbdata){
		global $db;
		if(!empty($dbdata)){
			if($this->debug != false){
				dump($dbdata['event_type']);
			}

			$dbdata['timestamp'] = time();
						
			$db->AutoExecute($this->LogTable, $dbdata, "INSERT");			
		}
	}

	/*
	**	Send an email to a profile's user
	*/
	function SendEmailAlarm($data, $eventlog){
		global $profile;

		$toUserMail = $data[4];
		$toUserName = $data[5];

		if($this->sandbox){
			$toUserMail = "mark@logaholic.com";
			$toUserName = "Mark Spier";
		}

		$fromTaName = $data[7];
		$fromTaEmail = $data[8];
		$ccMail = $data[9];	

		$notes = $data[10];

		$observed = $eventlog['ObserverdMetric'];
		$metric = $eventlog['metric'];
		$profilename = $eventlog['profile'];
		$timerange = date(trim(implode($profile->dateFormat)), $eventlog['start_date']) . " - " . date(trim(implode($profile->dateFormat)), $eventlog['end_date']);
		$xvisits = ($metric == "Visits") ? "" : "out of {$eventlog['ObserverdVisits']} visits";

		$xrange = array(0 , (int)number_format((1.2 * max($eventlog['baseline'], $eventlog['posteriorHigh'])) ,3));
		$json_xrange = json_encode($xrange);
				
		if(substr($eventlog['event_type'], 0,1) == 3){
			$event = "raised";
			$threshold = "Within threshold";
		} else {
			$event = "reset";
			$threshold = "Below threshold";
		}
		
		# Get the data for the graphs

		$distributionPlotData = $this->PlotDistributionData($eventlog);

		$historicalPlotData = array();
		$metricData = $this->PlotMetricData();

		$metricData['legend'][] = $this->LegendItem('Visits Per Hour','#68ADB5');
		$historicalPlotData[] = $metricData['Visits'];

		if($metric != "Visits" && !empty($metricData[$metric])){
			$metricData['legend'][] = $this->LegendItem(ucwords($metric) . ' Per Hour','#1F77B4');
			$historicalPlotData[] = $metricData[$metric];
		}

		require "templates/transalarmmail.php";

		$this->SendMail($toUserMail,$subject,$content,$graphdivs, $headers);
	}

	function PlotDistributionData($eventlog){

		$distribution_line = array();

		$obsv_hr = $eventlog['ObserverdVisits'] / $eventlog['hours'];
		$distymax = 0;		
		$yposteriorH = 0;
		$yposteriorL = 0;

		for ($i=1; $i <= 1000; $i++) { 
			$x = 0.2 * MIN($eventlog['baseline'], $obsv_hr) + $i / 1000 * (2*MIN($eventlog['baseline'], $obsv_hr) - 0.2 * MIN($eventlog['baseline'], $obsv_hr));
			
			$y = stats_dens_normal($x , $eventlog['mean'] , 2 );
			
			if(is_nan($y) || $y <= 0){ $y = 0; }

			if($y > $distymax){ $distymax = $y; }

			if(round($x) == round($eventlog['posteriorHigh'])){
				$eventlog['posteriorHigh'] = $x;
				$yposteriorH = $y;
			}
			if(round($x) == round($eventlog['posteriorLow']) && $yposteriorL === 0 ){
				$eventlog['posteriorLow'] = $x;
				$yposteriorL = $y;
			}
				
			$distribution_line[] = array($x,$y, null);
		}

		$mean_line = array( 
			array($eventlog['mean'],0, null),
			array($eventlog['mean'],$distymax, $eventlog['mean']) 
		);

		$Expected_line = array( 
			array($eventlog['baseline'],0, null),
			array($eventlog['baseline'],$distymax + 0.03, $eventlog['baseline']) 
		);
		$Threshold_line = array( 
			array($eventlog['ROPEboundary'],0, null),
			array($eventlog['ROPEboundary'],$distymax + 0.03, $eventlog['ROPEboundary']) 
		);

		$posteriorLow_line = array( 
			array($eventlog['posteriorLow'],0, null),
			array($eventlog['posteriorLow'],$yposteriorL, $eventlog['posteriorLow']) 
		);

		$posteriorHigh_line = array( 
			array($eventlog['posteriorHigh'],0, null),
			array($eventlog['posteriorHigh'],$yposteriorH, $eventlog['posteriorHigh']) 
		);

		$data = array(
			"legend" => array(
				$this->LegendItem("Expected",'#FFA14E'),
				$this->LegendItem("Threshold",'#FFBF80'),
				$this->LegendItem("Mean",'#98E3F9'),
				$this->LegendItem( number_format(($this->config['ErrorCertainty'] * 100), 1) . "% percentile" ,'#AEC7E8'),
				$this->LegendItem( number_format(((1 - $this->config['ResetCertainty']) * 100), 1) ."% percentile" ,'#68ADB5'),
				$this->LegendItem("Distribution",'#1F77B4'),
			),
			"data" => array(  $Expected_line, $Threshold_line,$mean_line, $posteriorHigh_line, $posteriorLow_line, $distribution_line )
		);


		return $data;
	}

	function LegendItem($name, $color){
		return "<div style='color: {$color}; float: left; clear:both;'>
			<div class='legend_bullet switchable'><div style='background-color: {$color};'></div></div><label style='line-height: 17px; margin-left:5px;'>{$name}</label>
		</div>";
	}

	function SendMail($recipients,$subject,$content,$graphdivs, $headers){

		$data = array(
			"recipients" => $recipients,
			"subject" => $subject,
			"content" => base64_encode($content),
			"from" => $headers[0],
			"to" => $headers[1],
			"cc" => $headers[2],
			"graph_0" => base64_encode($graphdivs[0]),
			"graph_1" => base64_encode($graphdivs[1])
		);
				
		$r = Post('http://software.logaholic.com/transalarm/alarm.php' , $data);

		echo $r;
	}

	function PlotMetricData(){
		global $profile;
		$data = array();

		$data['mindate'] = "";
		$data['legend'] = array();

		foreach ($this->processData as $row) {
			$k = $row[1];
			# preset as array
			if(empty($data[$k])){
				$data[$k] = array();
			}

			$date = date(trim(implode($profile->dateFormat)) . " H", $row[0]);

			if(empty($data['mindate'])){  $data['mindate'] = $date; }

			# add the values for the graph
			$data[$k][] = array( $date , (int)$row[2]);
		}

		return $data;
	}

	/*
	**	Lookup the detail data for profiles we need to email
	*/
	function EmailAlarms($log){
		$row = 0;

		if(empty($this->alarmProfiles)){
			return false;
		}

		$fp = fopen($this->transAlarmRecipientsFile , 'r');
		while (($data = fgetcsv($fp, 1000, ",")) !== FALSE) {
			if($row == 0){ $row ++; continue; }

			$profilename = $data[0];


			# Remove the '.' from the start of the profilename
			if(substr($profilename, 0, 1) == "."){
        		$profilename = substr($profilename, 1); 
			}
			$profilename = str_replace(" ","", $profilename);

			# Send alarm emails to the profile's user
			if(in_array($profilename, $this->alarmProfiles)){
				$this->SendEmailAlarm($data, $log);
				break;
			}
			$row ++;
		}
		fclose($fp);
		return true;
	}

	/*
	** Collect the ids of the kpis based on the profile's set targetfiles
	*/
	function KPIids(){
		global $db, $profile;
		$kpis = explode(",", $profile->targetfiles );	

		$query = "SELECT id, url FROM {$profile->tablename_urls} WHERE `url` IN ('". implode("','",$kpis) ."')";
		$r = $db->Execute($query);

		$data = array();
		while($row = $r->fetchRow()){
			$data[ $row['id'] ] = $row['url'];
		}
		return $data;
	}

	/*
	**	Set Data for the KPI's which we use for processing
	*/
	function GetKPIData(){
		global $db, $profile;

		if(empty($this->kpis)){
			return array();
		}

		$ids = array();
		foreach ($this->kpis as $id => $url) {
			$ids[] = $id;
		}		
		
		$query = "SELECT `timestamp`, u.url as metric ,count(DISTINCT visitorid) as hits ". 	
				"FROM $profile->tablename as a, $profile->tablename_urls as u ". 
				"WHERE `timestamp` >= ? AND `timestamp` <= ?  ".
				"AND a.url IN (". implode(",", $ids) . ") " .
				"AND a.url = u.id " .
				"GROUP BY FROM_UNIXTIME(`timestamp`, ? ) ,a.url ORDER BY timestamp ASC";

		$r = $db->Execute($query, array($this->from ,$this->to, '%d-%m-%Y-%H' ) );

		while($row = $r->fetchRow()){

			$this->processData[] = array($row[0],$row[1],$row[2]); 

			if(!isset($this->sumOfMetrics[ $row[1] ])){ 
				$this->sumOfMetrics[ $row[1] ] = 0; 
			}
			$this->sumOfMetrics[ $row[1] ] += $row[2];

		}
	}

	/*
	**	Set Data for the Visits and Transaction which we use for processing
	*/
	function GetMainData(){
		global $db, $profile;


		$query = "SELECT `timestamp`, COUNT(distinct sessionid) as visits ";

		$query .= ($this->transactions) ? ",count(DISTINCT transaction_id) as transactions " : "";		

		$query .= "FROM {$profile->tablename} ".
			"WHERE `timestamp` >= ? AND `timestamp` <= ? ".
			"GROUP BY FROM_UNIXTIME(`timestamp`, ? ) ORDER BY timestamp ASC";

		$r = $db->Execute($query, array($this->from ,$this->to, '%d-%m-%Y-%H' ) );
		
		$this->firstDataDateHour = 0;

		while($row = $r->fetchRow()){
			# Get the first timestamp in our data set
			if($this->firstDataDateHour === 0){
				$this->firstDataDateHour = $row[0];
			}

			$this->lastDataDateHour = $row[0];
			
			$this->processData[] = array($row[0],'Visits',$row[1]);
			$this->sumOfMetrics['Visits'] += $row[1];

			if($this->transactions){
				$this->processData[] = array($row[0],'Transactions',$row[2]);
				$this->sumOfMetrics['Transactions'] += $row[2];
			}
		}
	}

	/*
	**	Parse data 
	*/
	function ParseAllData(){
		global $db, $profile;

		# Get the id's for the KPI's
		$this->kpis = $this->KPIids();

		# Do we have any results in the TBL_TRANSALARM_DATA table ?
		$r = $db->Execute("SELECT * FROM ". $this->persistantDataTable . " WHERE profile = ? LIMIT 1", array($profile->profilename) );
		$result = $r->fetchRow();

		$this->GetProcessData();



		# If there is no data for this profile in the table then reset the baseline for this profile.
		if(empty($result)){
			$this->ResetBaseline();
		} else{
			$this->CheckPersistantMetrics();

			# We have data in the TBL_TRANSALARM_DATA table. Run the TestModule
			$log = $this->TestModule();

			if(!$this->sandbox){
				# Send Alarms when needed
				$this->EmailAlarms($log);
			}
		}
	}

	function CheckPersistantMetrics(){
		global $db, $profile;


		$metrics = $this->kpis;
		$metrics[] = 'Visits';

		if($this->transactions){
			$metrics[] = 'Transactions';
		}


		foreach ($metrics as $m) {
			$hash = md5($profile->profilename . $m);
			# Do we have any results in the TBL_TRANSALARM_DATA table ?
			$r = $db->Execute("SELECT * FROM ". $this->persistantDataTable . " WHERE hash = ? LIMIT 1", array($hash) );
			$result = $r->fetchRow();

			if(empty($result)){
				# add the data to the TBL_TRANSALARM_DATA table..
				$q = "INSERT INTO ". $this->persistantDataTable . "(`profile`, `metric`,`hash`) VALUES (?,?,?)";
				$db->Execute($q, array($profile->profilename, $m, $hash ) );

				# Add a reset log type to the Log table
				$this->LogEvent(array(
					"event_type" => $this->events[5],
					"profile" => $profile->profilename,
					"metric" => $m
				));
			}
		}
	}

	/*
	** Set the data we need to process
	*/
	function GetProcessData(){
		# Reset the processData array
		$this->processData = array();

		# Reset sum of metrics
		foreach ($this->sumOfMetrics as $m => $c) {
			$this->sumOfMetrics[$m] = 0;
		}

		# Set the main data to the processData array. (Visits and Transactions)
		$this->GetMainData();

		# Set the data for the KPI's metrics
		$this->GetKPIData();
	}

	/*
	** Reset the default baseline for existing profiles which do not have any data in TBL_TRANSALARM_DATA
	** optional : targetMetric = target metric we want to resset, used for reseting baseline for one metric after the test
	*/
	function ResetBaseline($targetMetric = ""){
		global $profile, $db;

		$grouped = array();

		# Add baseline for all metrics except Visits we do that later on..
		foreach ($this->sumOfMetrics as $metric => $v) {
			# Skip visits reset when we do a specific reset only
			if(!empty($targetMetric) && $targetMetric != $metric){
				continue;
			}

			if($metric != "Visits"){
				$baseline = $v / $this->sumOfMetrics['Visits'];
				$regressors = 'N/A';

				if($baseline == 0){
					if($metric == 'Transactions'){
						$baseline = $this->config['DefaultBaselines']['Transactions'];
					} else {
						$baseline = $this->config['DefaultBaselines']['[default]'];
					}
				}

				# Create a hash to make the profile and metric combination unique
				$hash = md5($profile->profilename . $metric);

				# add the data to the TBL_TRANSALARM_DATA table..
				$q = "INSERT INTO ". $this->persistantDataTable . "(`profile`, `metric`, `baseline`, `regressors`,`hash`) VALUES (?,?,?,?,?)";
				$q .= " ON DUPLICATE KEY UPDATE baseline = VALUES(baseline) , regressors = VALUES(regressors), `last0+` = VALUES(`last0+`)";
				$s = array($profile->profilename, $metric, $baseline , $regressors, $hash );
				$db->Execute($q, $s );

				# Add a reset log type to the Log table
				$this->LogEvent(array(
					"timestamp" => time(),
					"event_type" => $this->events[6], 
					"profile" => $profile->profilename,
					"metric" => $metric,
					"baseline" => $baseline
				));
			}
		}

		# Skip visits reset when we do a specific reset only
		if(!empty($targetMetric) && $targetMetric != 'Visits'){
			return;
		}

		# Create an array with days and hours as keys to sum the data for the regressors
		foreach ($this->processData as $k => $v) {
			if($v[1] != "Visits"){ continue; }

			$this->totalHours ++;
			
			$day = date('l', $v[0]);
			$hour = date('H', $v[0]);

			@$grouped[$day][ date('W', $v[0]) ] += $v[2];
			$grouped[$hour][] = $v[2];
		}

		# Set the baseline for the visits
		$baseline = $this->sumOfMetrics['Visits'] / ( $this->config['ResetDays'] * 24 );

		$regressors = array();
		foreach ($grouped as $key => $data) {
			if(is_numeric($key)){
				$regressors[$key] = $this->CalcRegrossorForHour($data);
			} else {
				
				$regressors[$key] = $this->CalcRegrossorForDay($data);
			}
		}

		# Create a hash to make the profile and metric combination unique
		$hash = md5($profile->profilename . 'Visits');

		# add the data to the TBL_TRANSALARM_DATA table if already exists..
				$q = "INSERT INTO ". $this->persistantDataTable . "(`profile`, `metric`, `baseline`, `regressors`,`hash`) VALUES (?,?,?,?,?)";
		$q .= " ON DUPLICATE KEY UPDATE baseline = VALUES(baseline) , regressors = VALUES(regressors), `last0+` = VALUES(`last0+`)";

		$s = array($profile->profilename, 'Visits', $baseline, json_encode($regressors), $hash );
		$db->Execute($q, $s );

		# Add a reset log type to the Log table
		$this->LogEvent(array(
			"event_type" => $this->events[6],
			"profile" => $profile->profilename,
			"metric" => 'Visits',
			"baseline" => $baseline
		));

	}

	/*
	**	Calculate the progressor for a hour
	**	Average(visits per hour for that hour of the day) / Average(visits per hour)
	*/
	function CalcRegrossorForHour($data){
		$avg_vph = $this->sumOfMetrics['Visits'] / ( $this->config['ResetDays'] * 24 );

		return ( (array_sum($data) / count($data) )  / $avg_vph );
	}

	/*
	**	Calculate the progressor for a day
	**	Average(visits per day for that day of the week) / Average(visits per day)
	*/
	function CalcRegrossorForDay($data){
		$avg_vpd = ( $this->sumOfMetrics['Visits'] / $this->config['ResetDays'] );

		return ( (array_sum($data) / count($data) )  / $avg_vpd );
	}


	/*
	** Get The Persistant Data for each metric of this profile
	*/
	function GetTaPersistantData($profilename){
		global $db;
		$r = $db->Execute("SELECT * FROM ". $this->persistantDataTable . " WHERE profile = ?", array($profilename) );
		$data = array();
		while($row = $r->fetchRow()){
			# Add the testStart to the row
			$row['testStart'] = max($row['last0+'], $row['last-'],$this->firstDataDateHour) + 3600;		

			# Add the testEnd to the row
			$row['testEnd'] = $this->lastDataDateHour;

			# If baseline is not set and the metric exists in DefaultBaselines, set baseline equal to DefaultBaselines
			if(!isset($this->config['DefaultBaselines'][ $row['metric'] ])){
				$row['baseline'] = (empty($row['baseline'])) ? $this->config['DefaultBaselines']['[default]'] : $row['baseline'];
			} else {
				$row['baseline'] = (empty($row['baseline'])) ? $this->config['DefaultBaselines'][ $row['metric'] ] : $row['baseline'];				
			}

			$row['observerdDenominator'] = 0;

			$data[ $row['metric'] ] = $row;
		}
		return $data;
	}


	/*
	** The Master Test Module
	*/	
	function TestModule(){
		global $profile, $db;

		# Get The Persistant Data for each metric of this profile
		$data = $this->GetTaPersistantData($profile->profilename);

		$sums = array();
		# Reset the sumOfMetrics array() to 0
		foreach ($this->sumOfMetrics as $m => $v) {
			$sums[$m] = 0; 
		}

		$hours = 0;

		# Here we loop through each hour of each metric
		foreach($this->processData as $k => $v){
			# Set the metric to $m
			$m = $v[1];

			# Sum between testStart and testEnd 
			if($v[0] >= $data[$m]['testStart'] && $v[0] <= $data[$m]['testEnd']){
				$sums[$m] += $v[2];

				# If metric=visits
				if($m == 'Visits'){
					$hours ++;
					$regressors = json_decode($data[$m]['regressors'] ,true);

					$hourOfDay = date("H", $v[0]);
					$dayOfWeek = date("l", $v[0]);

					# sum over each hour from testStart to testEnd inclusive
					# Σ( regressors[hourOfDay(h)] * regressors[dayOfWeek(h)] )
					if(!empty($regressors)){
						$data[$m]['observerdDenominator'] += $regressors[$hourOfDay] * $regressors[$dayOfWeek];
					}
				}
			}
		}



		foreach ($data as $m => $v) {
			# Set observerdMetric as Sum between testStart and testEnd 
			$observerdMetric = (empty($sums[$m])) ? 0 : $sums[$m];

			# If metric is not Visits, the number of visits from testStart to testEnd inclusive
			if($m != 'Visits'){
				$data[$m]['observerdDenominator'] = $sums['Visits'];
			}

			# Set the amount of hours between testStart and testEnd
			$test_hours = ceil(($v['testEnd'] - $v['testStart']) / 3600);

			# Set MetricROPEPercentage by looking up the current metric in config ROPEPercentage
			$MetricROPEPercentage = (isset($this->config['ROPEPercentage'][$m])) ? $this->config['ROPEPercentage'][$m] : $this->config['ROPEPercentage']['[default]'];

			# If metric is not Visits
			if($m != 'Visits'){
				# ROPEBoundry is baseline*(1-MetricROPEPercentage)
				$ROPEBoundry = $v['baseline']*(1-$MetricROPEPercentage);

				# alpha is 1+observedMetric+baseline
				$alpha = 1 + $observerdMetric + $v['baseline'];

				# beta is: observerdVisits - observerdMetric +(1 / (ROPEBoundry / ((1 + ROPEBoundry) - 1 / 3)) + ( 1 + ROPEBoundry)+ 2 / 3)
				$beta = $sums['Visits'] - $observerdMetric + (1 / ($ROPEBoundry / ((1 + $ROPEBoundry) - 1 / 3)) + ( 1 + $ROPEBoundry )+ 2 / 3);

				$posteriorHigh = Math::BETAINV($this->config['ErrorCertainty'] , $alpha , $beta );
				
				$posteriorLow = Math::BETAINV((1 - $this->config['ResetCertainty']) , $alpha , $beta );

				$mean = $alpha / ($alpha + $beta);

			# If metric=visits
			} else {
				# Set empty variables if needed
				if(!empty($observerdMetric)){
					# ROPEBoundry is: baseline *(1-MetricROPEPercentage)* ((count of hours from testStart to testEnd) / observerdDenominator)
					$ROPEBoundry = $v['baseline'] *(1-$MetricROPEPercentage)* (($test_hours) / $data[$m]['observerdDenominator']);

					# alpha is GammaAlphaConstant + (observedMetric / ROPEBoundry)
					# beta is (1+observerdDenominator)/ROPEBoundry
					if(empty($ROPEBoundry)){
						$alpha = $this->GammaAlphaConstant;
						$beta = 0;
						$posteriorHigh = 0;
						$posteriorLow = 0;
						$mean = 0;
					} else{
						$alpha = $this->GammaAlphaConstant + ($observerdMetric / $ROPEBoundry);
						// $beta = (1 + $data[$m]['observerdDenominator']) / $ROPEBoundry;
						$beta = (1 + $hours) / $ROPEBoundry;

						$posteriorHigh = stats_cdf_gamma($this->config['ErrorCertainty'] , $alpha , $beta , 2);	

						$posteriorLow  = stats_cdf_gamma((1 - $this->config['ResetCertainty']) , $alpha , $beta , 2 );

						$mean = $alpha / $beta;

					}
				} else {					
					$ROPEBoundry = 0;
					$alpha = $this->GammaAlphaConstant;
					$beta = 0;
					$posteriorHigh = 0;
					$posteriorLow = 0;
					$mean = 0;
				}
			}	

			$logData = array(
				"timestamp" => 0,
				"event_type" => null,
				"profile" => $profile->profilename,
				"metric" => $m,
				"start_date" => $v['testStart'],
				"end_date" => $v['testEnd'],
				"baseline" => $v['baseline'],
				"alpha" => $alpha,
				"beta" => $beta,
				"mean" => $mean,
				"hours" => $hours,
				"ROPEboundary" => $ROPEBoundry,
				"ObserverdDenominator" => $data[$m]['observerdDenominator'],
				"ExpectedMetric" => null,
				"ObserverdMetric" => $observerdMetric,
				"ObserverdVisits" => $sums['Visits'],
				"ErrorCertainty" => $this->config['ErrorCertainty'],
				"posteriorHigh" => (is_nan($posteriorHigh) ? null : $posteriorHigh),
				"ResetCertainty" => $this->config['ResetCertainty'],
				"posteriorLow" => (is_nan($posteriorLow) ? null : $posteriorLow),
			);

			# Display debug output
			$this->DebugOutput($logData,$alpha,$beta,$m);		

			if($v['last0+'] == ($v['testStart'] - 3600)){
				$last_state = '0+';
			} else if($v['last-'] == ($v['testStart'] - 3600)){
				$last_state = '-';
			} else {
				$last_state = '?';
			}

			# If $posteriorHigh < $ROPEBoundry 
			if($posteriorHigh < $ROPEBoundry){
				# Then follow path -

				if($last_state == '-'){
					#Log "2 Ongoing alarm"  Event
					$logData["event_type"] = $this->events[2];
					$this->LogEvent($logData);

					# check if we need to reset and finish - path
					$this->Finish($v['last0+'] , $m);
				} else {
					# Log "1 Alarm fired"  Event
					$logData["event_type"] = $this->events[1];
					$this->LogEvent($logData);

					# Trigger Email
					$this->alarmProfiles[] = $profile->profilename;
				}


				# Update last- in persistent storage to lastDataDateHour
				$q = "UPDATE ". $this->persistantDataTable . " SET `last-` = ? WHERE hash = ?";
				$db->Execute($q , array( $this->lastDataDateHour, md5($profile->profilename . $m) ) );

			} elseif ($posteriorLow > $ROPEBoundry) {
				# Then follow path 0+
				
				if($last_state == '0+'){
					# Log "4 All ok" Event
					$logData["event_type"] = $this->events[4];
					$this->LogEvent($logData);

					# check if we need to reset and finish - path
					$this->Finish($v['last-'] , $m);
				} else {
					# Log "3 Alarm active" Event
					$logData["event_type"] = $this->events[3];
					$this->LogEvent($logData);

					# Trigger Email
					$this->alarmProfiles[] = $profile->profilename;
				}


				# Update last0+ in persistent storage to lastDataDateHour
				$q = "UPDATE ". $this->persistantDataTable . " SET `last0+` = ? WHERE hash = ?";
				$db->Execute($q , array( $this->lastDataDateHour, md5($profile->profilename . $m) ) );

			} else {
				# Otherwise follow path ?
				# Log "7 Inclusive" Event
				$logData["event_type"] = $this->events[7];
				$this->LogEvent($logData);
			}

			return $logData;
		}
	}

	function Finish($check, $metric){
		# Any [check] events in last R days?
		if($check >= $this->firstDataDateHour && $check <= $this->lastDataDateHour){
			# Finish
		} else {
			# ResetBaseline
			$this->ResetBaseline($metric);
		}
	}

	function DebugOutput($data,$a, $b, $metric){
		if($this->debug == false){
			return;
		}

		echo "<hr color='silver'/>";
		dump("Metric: ".$metric);

		dump("Test Start: ".  date("Y-m-d H", $data['start_date'] ) );
		dump("Test End: ".  date("Y-m-d H", $data['end_date'] ) );


		if($metric == "Visits"){
			dump("# Set observerdMetric as Sum between testStart and testEnd");		
			dump("observerdMetric: ". $data['ObserverdMetric']);	

			dump("ObserverdDenominator: # &Sigma;( regressors[hourOfDay(h)] * regressors[dayOfWeek(h)] )");
			dump("ObserverdDenominator: ". $data['ObserverdDenominator']);		

			dump("Baseline: ".  $data['baseline'] );

			dump("# ROPEBoundry is: baseline *(1-MetricROPEPercentage)* ((count of hours from testStart to testEnd) / observerdDenominator)");		
			dump("ROPEBoundry: ". $data['ROPEboundary']);	

			dump("# alpha is GammaAlphaConstant + (observedMetric / ROPEBoundry)");		
			dump("alpha: ".$a);		

			dump("# beta is (1+observerdDenominator)/ROPEBoundry");
			dump("beta: ".$b);

		} else {	

			dump("# Set observerdMetric as Sum between testStart and testEnd");		
			dump("observerdMetric: ". $data['ObserverdMetric']);	

			dump("# ObserverdDenominator is the number of visits from testStart to testEnd inclusive");
			dump("ObserverdDenominator: ". $data['ObserverdDenominator']);

			dump("Baseline: ".  $data['baseline'] );

			dump("# ROPEBoundry is baseline*(1-MetricROPEPercentage)");		
			dump("ROPEBoundry: ". $data['ROPEboundary']);	

			dump("# alpha is 1+observedMetric+baseline");		
			dump("alpha: ".$a);		

			dump("# beta is: observerdVisits - observerdMetric +(1 / (ROPEBoundry / ((1 + ROPEBoundry) - 1 / 3)) + ( 1 + ROPEBoundry)+ 2 / 3)");
			dump("beta: ".$b);
		}

		dump("posteriorHigh: ".$data['posteriorHigh']);
		dump("posteriorLow: ". $data['posteriorLow']);		
	}
    
}
$logaTA = new TransAlarm();
?>